@@ -5,7 +5,7 @@ use strict;
 use warnings;
 use autouse 'Carp' => qw(carp croak confess);
 
-our $VERSION = 0.12;
+our $VERSION = 0.13;
 
 =head1 NAME
 
@@ -38,18 +38,20 @@ our @EXPORT    = qw( );
 
 # Exportable subroutine definitions
 #
+sub configure( % );
 sub date_of( $ );
 sub get_IP_address ( $ );
 sub get_interfaces_href();
 sub get_interfaces_list();
 sub get_trace_summary_href();
-sub process_trace( $ ; $$ );
+sub process_trace( $ );
 sub records_in( $ );
 sub verbose();
 sub write_interface_summaries( ; $);
 sub write_trace_summary( ; $ );
 
 our @EXPORT_OK = qw(
+		    configure
 		    date_of
 		    get_IP_address
 		    get_interfaces_href
@@ -82,6 +84,8 @@ sub progress( $ );
 sub write_summary( *$ ; $ );
 sub print_value( *$ );
 
+our %options;
+
 # Load the IANA protocol numbers from the __DATA__ section.  If by any
 # chance we end up having duplicate keywords, something must have
 # corrupted the __DATA__ section, so abort.
@@ -100,6 +104,21 @@ INIT {
 
     $iana_protocol_numbers{$k} = $v;
   }
+
+  # Default options, parameters and output
+  #
+  %options = (
+	      # Do not display progress information
+	      Verbosity => 0,
+
+	      'Link Capacity' => 155_520_000, # Bits per second
+
+	      # Filename to store TCP traffic in tcpdump format
+	      tcpdump => 0,
+
+	      # Filename to store TCP traffic in ns2 format
+	      ns2 => 0,
+	     );
 }
 
 # Used to sort the keys of a hash in numeric order instead of the
@@ -108,10 +127,6 @@ INIT {
 #
 sub numerically { $a <=> $b; }
 
-# By default, assume the user does not want any progress information.
-#
-my $Verbose = 0;
-
 =head1 INSTALLATION
 
 C<Net::Traces::TSH> can be installed like any CPAN module.  In
@@ -221,7 +236,7 @@ A single TSH trace may contain records for packets observed on several
 different interfaces.  For example, the daily TSH traces from the
 NLANR PMA repository typically contain records from two different
 interfaces. In such cases, incoming and outgoing traffic can be
-differentiated based on the interface number (despite the scrabling of
+differentiated based on the interface number (despite the scrabbling of
 IP addresses to protect privacy).  C<Net::Traces::TSH> users may be
 interested in collecting statistical information for each interface
 separately or aggregating across the entire trace.
@@ -306,7 +321,7 @@ C<undef> if traffic directionality was not examined.
 
 =item $Trace{'Link Capacity'}
 
-The L<capacity of the monitored link|"process_trace"> in bits per
+The L<capacity of the monitored link|"configure"> in bits per
 second (b/s).
 
 =back
@@ -498,6 +513,48 @@ C<Net::Traces::TSH> does not export any functions by default.  The
 following functions, listed in alphabetical order, are
 L<exportable|"EXPORTS">.
 
+=head2 configure
+
+  configure %OPTIONS
+
+Used to specify verbosity, the link capacity, and the types of outputs
+requested.  For example,
+
+ configure(
+           # Display progress information, equivalent to calling verbose()
+           #
+           Verbosity       => 1, # default is 0, no progress information
+
+           'Link Capacity' => 100_000_000, # bits per second
+
+           # Convert the TCP records in the TSH trace to tcpdump
+           # format and store in 'trace.tcpdump'.
+           #
+           tcpdump         => 'trace.tcpdump',
+
+           # Convert the TCP data-carrying segment records to binary
+           # ns2 traffic trace format.  Create one binary file per
+           # interface and use 'trace.ns2' as the file prefix.
+           #
+           ns2         => 'trace.ns2',
+
+          );
+
+=cut
+
+sub configure ( % ) {
+  while ( defined ($_ = shift) ) {
+    if ( defined $options{$_} ) {
+      $options{$_} = shift;
+    }
+    else {
+      $options{$_} = undef;
+      shift;
+      carp "Unknown configuration option '$_' ignored";
+    }
+  }
+}
+
 =head2 date_of
 
   date_of FILENAME
@@ -551,7 +608,9 @@ Returns a hash I<reference> to L<%Interfaces|"Data Structures">.
 =cut
 
 sub get_interfaces_href() {
+
   return \%Interfaces;
+
 }
 
 
@@ -566,8 +625,10 @@ in the trace.
 =cut
 
 sub get_interfaces_list() {
+
   return wantarray ? sort numerically keys %Interfaces
                    : scalar keys %Interfaces;
+
 }
 
 =head2 get_trace_summary_href
@@ -579,57 +640,23 @@ Returns a hash I<reference> to L<%Trace|"Data Structures">.
 =cut
 
 sub get_trace_summary_href() {
+
   return \%Trace;
+
 }
 
 
 =head2 process_trace
 
  process_trace FILENAME
- process_trace FILENAME, TCPDUMP_OUTPUT
- process_trace FILENAME, TCPDUMP_OUTPUT, NUMBER
 
-If called in a void context process_trace() examines the binary TSH
-trace stored in FILENAME, and populates L<%Trace and %Interfaces|"Data
+In a void context, process_trace() examines the binary TSH trace stored
+in FILENAME, and populates L<%Trace and %Interfaces|"Data
 Structures">.
 
-If TCPDUMP_OUTPUT is specified, process_trace() generates a text file
-based on the trace records in a format similar to the modified output
-of F<tcpdump>, as presented in I<TCP/IP Illustrated Volume 1> by
-W. R. Stevens (see pp. 230-231).
-
-You can use such an output as input to other tools, present real
-traffic scenarios in a classroom, or simply "eyeball" the trace.  For
-example, here are the first ten lines of the contents of such a file:
-
- 0.000000000 10.0.0.1.6699 > 10.0.0.2.55309: . ack 225051666 win 65463
- 0.000014000 10.0.0.3.80 > 10.0.0.4.14401: S 457330477:457330477(0) ack 810547499 win 34932
- 0.000014000 10.0.0.1.6699 > 10.0.0.2.55309: . 3069529864:3069531324(1460) ack 225051666 win 65463
- 0.000024000 10.0.0.5.12119 > 10.0.0.6.80: F 2073668891:2073668891(0) ack 183269290 win 64240
- 0.000034000 10.0.0.7.4725 > 10.0.0.8.445: S 3152140131:3152140131(0) win 16384
- 0.000067000 10.0.0.1.6699 > 10.0.0.2.55309: P 3069531324:3069531944(620) ack 225051666 win 65463
- 0.000072000 10.0.0.11.3381 > 10.0.0.12.445: S 1378088462:1378088462(0) win 16384
- 0.000083000 10.0.0.13.1653 > 10.0.0.1.6699: P 3272208349:3272208357(8) ack 501563814 win 32767
- 0.000093000 10.0.0.14.1320 > 10.0.0.15.445: S 3127123478:3127123478(0) win 64170
- 0.000095000 10.0.0.4.14401 > 10.0.0.3.80: R 810547499:810547499(0) ack 457330478 win 34932
-
-Note that this output is similar to what F<tcpdump> with options C<-n>
-and C<-S> would have produced.  The only missing fields are related to
-the TCP options negotiated during connection setup.  Unfortunately,
-L<TSH records|"DESCRIPTION"> include only the first 16 bytes of the
-TCP header, making it impossible to record the options from the
-segment header.
-
-NUMBER specifies the L<capacity of the monitored link|"General Trace
-Information"> in bits per second (b/s).  If not specified, it defaults
-to 155,520,000.
-
-=head3 Context matters
-
-When called in a list context process_trace() in addition to
-collecting summary statistics, it extracts all TCP flows and TCP
-data-carrying segments from the trace, returning two hash references.
-For example,
+In a list context process_trace() in addition to collecting summary
+statistics, it extracts all TCP flows and TCP data-carrying segments
+from the trace, returning two hash references.  For example,
 
  my ($senders_href, $segments_href) = process_trace 'trace.tsh';
 
@@ -702,6 +729,80 @@ CAVEAT: write_sojourn_times() is not currently included in the stable,
 CPAN version of the module.  L<Contact me|"AUTHOR"> if you want to get
 a copy of the bleeding edge version.
 
+=head3 Using a TSH trace in ns2 simulations
+
+In addition to extracting %senders and %segments, C<Net::Traces::TSH>
+allows you to generate binary files suitable for driving L<ns2
+simulations|"SEE ALSO">.  For example,
+
+ configure(ns2 => 'some.tsh');
+
+ process_trace 'some.tsh';
+
+After the call to configure(), process_trace() will generate a binary
+file for each interface found in the trace.  For example, assume that
+F<some.tsh> has recorded traffic from two interfaces, 1 and 2.
+process_trace() will generate two binary files:
+
+  some.tsh-if1.bin
+  some.tsh-if2.bin
+
+Each of these files L<can be used in ns2 simulations|"SEE ALSO"> in
+conjunction Application/Traffic/Trace. For example, the following ns2
+script fragment illustrates how to attach F<some.tsh-if2.bin> to a
+traffic source
+
+ # ...
+
+ # Initialize a trace file
+ #
+ set tfile [new Tracefile]
+ $tfile filename some.tsh-2.bin
+
+ # Attach the tracefile
+ #
+ set trace [new Application/Traffic/Trace]
+ $trace attach-tracefile $tfile
+
+ # ...
+
+
+
+
+=head3 Converting TSH to F<tcpdump>
+
+If you would like to extract the TCP traffic and store it in
+F<tcpdump> format, use
+
+ configure(tcpdump => 'tcpdump_filename');
+
+before calling process_trace(). process_trace() will generates a text
+file based on the trace records in a format similar to the modified
+output of F<tcpdump>, as presented in I<TCP/IP Illustrated Volume 1>
+by W. R. Stevens (see pp. 230-231).
+
+You can use such an output as input to other tools, present real
+traffic scenarios in a classroom, or simply "eyeball" the trace.  For
+example, here are the first ten lines of the contents of such a file:
+
+ 0.000000000 10.0.0.1.6699 > 10.0.0.2.55309: . ack 225051666 win 65463
+ 0.000014000 10.0.0.3.80 > 10.0.0.4.14401: S 457330477:457330477(0) ack 810547499 win 34932
+ 0.000014000 10.0.0.1.6699 > 10.0.0.2.55309: . 3069529864:3069531324(1460) ack 225051666 win 65463
+ 0.000024000 10.0.0.5.12119 > 10.0.0.6.80: F 2073668891:2073668891(0) ack 183269290 win 64240
+ 0.000034000 10.0.0.7.4725 > 10.0.0.8.445: S 3152140131:3152140131(0) win 16384
+ 0.000067000 10.0.0.1.6699 > 10.0.0.2.55309: P 3069531324:3069531944(620) ack 225051666 win 65463
+ 0.000072000 10.0.0.11.3381 > 10.0.0.12.445: S 1378088462:1378088462(0) win 16384
+ 0.000083000 10.0.0.13.1653 > 10.0.0.1.6699: P 3272208349:3272208357(8) ack 501563814 win 32767
+ 0.000093000 10.0.0.14.1320 > 10.0.0.15.445: S 3127123478:3127123478(0) win 64170
+ 0.000095000 10.0.0.4.14401 > 10.0.0.3.80: R 810547499:810547499(0) ack 457330478 win 34932
+
+Note that this output is similar to what F<tcpdump> with options C<-n>
+and C<-S> would have produced.  The only missing fields are related to
+the TCP options negotiated during connection setup.  Unfortunately,
+L<TSH records|"DESCRIPTION"> include only the first 16 bytes of the
+TCP header, making it impossible to record the options from the
+segment header.
+
 =cut
 
 # A TSH record is 44 bytes long.
@@ -713,7 +814,7 @@ use constant TSH_RECORD_LENGTH => 44;
 #
 use constant TIMESTAMP_COLLISION_THRESHOLD => 3;
 
-sub process_trace( $ ; $$ ) {
+sub process_trace( $ ) {
 
   # Sanity checks
   #
@@ -731,11 +832,13 @@ sub process_trace( $ ; $$ ) {
 
   binmode INPUT; # Needed for non-UNIX OSes; no harm in UNIX
 
-  my $text_trace_filename = shift;
+  $options{tcpdump} and
+    ( open(TCPDUMP, '>', $options{tcpdump})
+      or croak "Cannot open $options{tcpdump}. $!"
+    );
 
-  $text_trace_filename and
-    ( open(TCPDUMP, '>', $text_trace_filename)
-      or croak "Cannot open $text_trace_filename. $!" );
+  my %ns2_fh;
+  my %ns2_previous_timestamp;
 
   progress "Initializing data structures... ";
 
@@ -743,7 +846,7 @@ sub process_trace( $ ; $$ ) {
 
   $Trace{filename} = $trace;
   $Trace{records} = $records;
-  $Trace{'Link Capacity'} = shift || 155_520_000;
+  $Trace{'Link Capacity'} = $options{'Link Capacity'};
 
   # If process_trace() is called in a void context, we will not
   # examine traffic direction, thus undef $Trace{unidirectional}.
@@ -1056,66 +1159,101 @@ sub process_trace( $ ; $$ ) {
 	}
       }
 
+      ##################################################################
+      #                Export %senders and %segments
+      ##################################################################
       # Determine if we should collect statistics about the %senders and
       # the %segments.  If process_trace() was called in a void context
       # then we do not need to collect such data, which results in
       # tremendous memory usage savings.
       #
-      if ( wantarray and $tcp_payload > 0 ) {
-	# Add elements to the hashes ONLY if the segment carries some
-	# payload.  This way, one can be more sure if a given segment
-	# was retransmitted or not, since ACKs are not guaranteed
-	# reliable delivery.
-	#
-	# Occasionally, we may get 2 or more TCP segments with the
-	# same $timestamp.  We would like to keep them in the segments
-	# hash and be able to discriminate between the different
-	# segments, so we use the following (hash) collision avoidance
-	# mechanism.
-	#
-	my $collisions = 0;
-
-	while ( exists $segments{$if}{$timestamp}{bytes} ) {
-	  # Sanity check: If more than TIMESTAMP_COLLISION_THRESHOLD
-	  # trace records have the same timestamp, it is better to
-	  # abort processing.  Theoretically there shouldn't be two
-	  # segments with the same timestamp.
+      if ( $tcp_payload > 0 ) {
+	if ( wantarray ) {
+	  # Add elements to the hashes ONLY if the segment carries
+	  # some payload.  This way, one can be more sure if a given
+	  # segment was retransmitted or not, since ACKs are not
+	  # guaranteed reliable delivery.
 	  #
-	  croak 'Too many duplicate timestamps: ', $collisions,
-	    ' trace records have the same timestamp. Processing aborted'
+	  # Occasionally, we may get 2 or more TCP segments with the
+	  # same $timestamp.  We would like to keep them in the
+	  # segments hash and be able to discriminate between the
+	  # different segments, so we use the following (hash)
+	  # collision avoidance mechanism.
+	  #
+	  my $collisions = 0;
+
+	  while ( exists $segments{$if}{$timestamp}{bytes} ) {
+	    # Sanity check: If more than TIMESTAMP_COLLISION_THRESHOLD
+	    # trace records have the same timestamp, it is better to
+	    # abort processing.  Theoretically there shouldn't be two
+	    # segments with the same timestamp.
+	    #
+	    croak 'Too many duplicate timestamps: ', $collisions,
+	          ' trace records have the same timestamp. Processing aborted'
 	    if $collisions++ == TIMESTAMP_COLLISION_THRESHOLD;
 
-	  carp "Duplicate timestamp $timestamp detected & replaced with ",
-	    $timestamp .= "1";
-	  $Trace{Transport}{TCP}{'Concurrent Segments'}++;
-	}
+	    carp "Duplicate timestamp $timestamp detected & replaced with ",
+	         $timestamp .= "1";
 
-	# Store the total length of the segment (headers + application
-	# payload), and the sequence number it carries
-	#
-	$segments{$if}{$timestamp}{bytes} = $ip_len;
-	$segments{$if}{$timestamp}{seq_num} = $seq_num;
+	    $Trace{Transport}{TCP}{'Concurrent Segments'}++;
+	  }
 
-	# In addition, flag by default every segment as an original
-	# transmission.  Detection of retransmitted segments is not
-	# done in process_trace(), but rather in write_sojourn_times()
-	#
-	$segments{$if}{$timestamp}{retransmitted} = undef;
+	  # Store the total length of the segment (headers +
+	  # application payload), and the sequence number it carries
+	  #
+	  $segments{$if}{$timestamp}{bytes} = $ip_len;
+	  $segments{$if}{$timestamp}{seq_num} = $seq_num;
 
-	# Add the packet timestamp to the respective sender list
-	#
-	push @{ $senders{$if}{"$src,$src_port,$dst,$dst_port"} },
+	  # In addition, flag by default every segment as an original
+	  # transmission.  Detection of retransmitted segments is not
+	  # done in process_trace(), but rather in
+	  # write_sojourn_times()
+	  #
+	  $segments{$if}{$timestamp}{retransmitted} = undef;
+
+	  # Add the packet timestamp to the respective sender list
+	  #
+	  push @{ $senders{$if}{"$src,$src_port,$dst,$dst_port"} },
 	     $timestamp;
 
-	# Flag bidirectional traffic found in the *same* interface.
-	# If bidirectional traffic is present in the same interface,
-	# it is not clear (yet) how to isolate "incoming" from
-	# "outgoing" traffic.
+	  # Flag bidirectional traffic found in the *same* interface.
+	  # If bidirectional traffic is present in the same interface,
+	  # it is not clear (yet) how to isolate "incoming" from
+	  # "outgoing" traffic.
+	  #
+	  $Trace{unidirectional} = 0
+	    if ( $Trace{unidirectional} and
+		 exists $senders{$if}{"$dst,$dst_port,$src,$src_port"}
+	       );
+	}
+
+	##################################################################
+	#        Export TSH to ns2 binary traffic trace format
+	##################################################################
+	# Generate an ns2 binary traffic trace. (TCP data-carrying
+	# segements only)
 	#
-	$Trace{unidirectional} = 0
-	  if ( $Trace{unidirectional} and
-	       exists $senders{$if}{"$dst,$dst_port,$src,$src_port"}
-	     );
+	if ( $options{ns2} ) {
+
+	  unless ( defined $ns2_fh{$if}) {
+	    open($ns2_fh{$if}, '>', "$options{ns2}-if$if.bin")
+	      or croak "Cannot open $options{ns2}-if$if.bin. $!";
+	    binmode $ns2_fh{$if}; # Needed for non-UNIX OSes; no harm in UNIX
+
+	    $ns2_previous_timestamp{$if} = $timestamp;
+	  }
+
+	  print
+	    { $ns2_fh{$if} }
+	    pack('NN', # two integers: interpacket time (usec), packet size (B)
+		 sprintf("%.0f", ( $timestamp
+				   - $ns2_previous_timestamp{$if} ) * 1_000_000
+			),
+		 $ip_len
+		);
+
+	  $ns2_previous_timestamp{$if} = $timestamp;
+	}
       }
 
       ##################################################################
@@ -1124,7 +1262,7 @@ sub process_trace( $ ; $$ ) {
       # Print a tcpdump-like time line of the TSH trace (for TCP
       # segments only)
       #
-      if ( $text_trace_filename ) {
+      if ( $options{tcpdump}) {
 	printf TCPDUMP "%1.9f ", $timestamp;
 	print TCPDUMP
 	  get_IP_address $src, ".$src_port > ",
@@ -1158,12 +1296,9 @@ sub process_trace( $ ; $$ ) {
 
   close INPUT;
 
-  if ( $text_trace_filename ) {
-    close TCPDUMP;
-
-    print STDERR "TCP activity stored in text format in $text_trace_filename\n"
-      if $Verbose;
-  }
+  close TCPDUMP and
+    progress "TCP activity stored in text format in $options{tcpdump}\n"
+  if $options{tcpdump};
 
   carp $Trace{Transport}{TCP}{'Concurrent Segments'},
     ' TCP segments had the same timestamp with another segment'
@@ -1261,7 +1396,7 @@ sub process_trace( $ ; $$ ) {
 }
 
 sub progress( $ ) {
-  print STDERR shift if $Verbose;
+  print STDERR shift if $options{Verbosity};
 }
 
 =head2 records_in
@@ -1291,12 +1426,21 @@ module.  By default C<Net::Traces::TSH> remains "silent".  Call
 verbose() to see trace processing progress indicators on standard
 error.
 
+As of version 0.13, verbose() is equivalent to
+
+ configure(Verbosity => 1);
+
 =cut
 
 sub verbose () {
-  $Verbose = 1;
+
+  $options{Verbosity} = 1;
+
 }
 
+# Utility function to export the information stored in %Trace and
+# %Interfaces in CSV format
+#
 sub write_summary( *$ ; $ ) {
   my ( $FH, $href, $if ) = @_;
 
@@ -1559,9 +1703,9 @@ None by default.
 
 =head2 Exportable
 
-date_of() get_IP_address() get_interfaces_href() get_interfaces_list()
-get_trace_summary_href() numerically() process_trace() records_in()
-verbose() write_trace_summary()
+configure() date_of() get_IP_address() get_interfaces_href()
+get_interfaces_list() get_trace_summary_href() numerically()
+process_trace() records_in() verbose() write_trace_summary()
 
 In addition, the following export tags are defined:
 
@@ -1584,7 +1728,7 @@ Finally, all exportable functions can be imported with
 
 =head1 VERSION
 
-This is C<Net::Traces::TSH> version 0.12.
+This is C<Net::Traces::TSH> version 0.13.
 
 =head1 SEE ALSO
 
@@ -1599,10 +1743,9 @@ at http://pma.nlanr.net/Traces .  The site contains a variety of
 traces gathered from several monitoring points at university campuses
 and (Giga)PoPs connected to a variety of large and small networks.
 
-
 C<Net::Traces::TSH> version 0.11 was presented in YAPC::NA 2004.  The
 presentation slides are available at
-http://www.cs.sunysb.edu/~kostas/art/yapc
+http://www.cs.stonybrook.edu/~kostas/art/yapc .
 
 =head2 DiffServ
 
@@ -1627,6 +1770,12 @@ K. K. Ramakrishnan I<et al.>, I<The Addition of Explicit Congestion
 Notification (ECN) to IP>, RFC 3168. Available at
 http://www.ietf.org/rfc/rfc3168.txt
 
+=head2 The ns2 network simulator
+
+C<Net::Traces::TSH> can convert TSH traces to binary files suitable to
+drive simulations in ns2. More information about ns2 is available at
+http://www.isi.edu/nsnam/ns/
+
 =head1 AUTHOR
 
 Kostas Pentikousis, kostas@cpan.org.
